#!/usr/bin/env python
# -*- coding:utf-8 -*-

# Program: 分库
# Author : HH
# Date   : 2016-02-03

import sys
import mysql.connector
import json
import copy

reload(sys)
sys.setdefaultencoding('utf-8')

class ShardingDatabase(object):
  '''
  创建分库类
  '''
  def __init__(self):
    # 设置默认的公共数据库链接参数
    self.db_config_common = {
      'user'    : 'root',
      'password': 'root',
      'host'    : '127.0.0.1',
      'port'    : 3306,
      'database': 'test'
    }
    self.db_config_from = copy.deepcopy(self.db_config_common)
    self.db_config_to   = copy.deepcopy(self.db_config_common)

  def get_conn_cursor(self, db_config = {}, source_type = 'common'):
    '''
    获取原数据库链接, 和游标
    '''
    if db_config and source_type == 'common':
      # 创建公共数据源连接
      self.db_config_common = db_config
      self.conn_common, self.cursor_select_common, self.cursor_dml_common \
        = self.create_conn_cursor(db_config)
      return self.conn_common, self.cursor_select_common, self.cursor_dml_common
    elif db_config and source_type == 'from':
      # 创建用户原数据源连接
      self.db_config_from = db_config
      self.conn_from, self.cursor_select_from, self.cursor_dml_from \
        = self.create_conn_cursor(db_config)
      return self.conn_from, self.cursor_select_from, self.cursor_dml_from
    elif db_config and source_type == 'to':
      # 创建用户目标数据源连接
      self.db_config_to = db_config
      self.conn_to, self.cursor_select_to, self.cursor_dml_to \
        = self.create_conn_cursor(db_config)
      return self.conn_to, self.cursor_select_to, self.cursor_dml_to

  def create_conn_cursor(self, db_config = {}):
    '''
    创建数据库连接并返回
    '''
    conn = mysql.connector.connect(**db_config)
    cursor_select = conn.cursor(buffered=True)
    cursor_dml = conn.cursor(buffered=True)
    return conn, cursor_select, cursor_dml

  def commit(self, source_type = 'common'):
    '''
    提交事物
    '''
    commit_code = \
    '''self.conn_{source_type}.commit()
    '''.format(source_type = source_type)
    exec commit_code

  def close_cursor(self, cursor_type, source_type = 'common'):
    '''
    关闭游标
    '''
    close_cursor_code = \
    '''self.cursor_{cursor_type}_{source_type}.close()
    '''.format(source_type = source_type,
               cursor_type = cursor_type)
    exec close_cursor_code

  def close_conn(self, source_type = 'common'):
    '''
    关闭链接
    '''
    close_conn_code = \
    '''self.conn_{source_type}.close()
    '''.format(source_type = source_type)
    exec close_conn_code

  def create_select_sql(self, table_name, cloumn_list=[], where_dict={}):
    '''
    通过给出的表明、列名、谓词 拼出sql
    '''
    if not cloumn_list:
      cloumn_list[0] = '*'
    cloumns = ', '.join(cloumn_list)
    sql = '''
      SELECT {cloumns}
      FROM {table_name}
      WHERE 1=1
    '''.format(cloumns    = cloumns,
               table_name = table_name)
    # 将where条件的字典转化为
    where_list = []
    for item in where_dict.iteritems():
      key = item[0]
      value = item[1]
      if not value:
        continue
      key_value = ''
      # 生成 key=value 形式
      if isinstance(value, str):
        key_value = '''{key} = '{value}' '''.format(
                                               key = key,
                                               value = value
                                             )
      else:
        key_value = '''{key} = {value}'''.format(
                                            key = key,
                                            value = value
                                          )
      where_list.append(key_value)
      
    # 构造 where 后面的字符串如： and name='value'
    if where_list:
      where_str = ' and '.join(where_list)
      sql += ' and ' + where_str
    return sql

  def create_insert_sql(self, table_name, columns, values):
    '''
    根据给出的列和值拼出INSERT语句
    '''
    if columns and values:
      value_list = []
      for value in values:
        if isinstance(value, unicode) or isinstance(value, str):
          value_list.append(''' '{value}' '''.format(value=value))
        else:
          value_list.append(str(value))

      columns_str = ', '.join(columns)
      values_str = ', '.join(value_list)
      insert_sql = '''
        INSERT INTO {table_name}({columns}) VALUES({values})
      '''.format(table_name = table_name,
                 columns    = columns_str,
                 values     = values_str)
      return insert_sql

  def create_update_sql(self, table_name, set_dict={}, where_dict={}):
    '''
    通过给出条件生成update语句
    '''
    if table_name and set_dict and where_dict:
      set_list = []
      where_list = []
      # 构造 set a = 'b'
      for key, value in set_dict.iteritems():
        if not value:
          continue
        key_value = ''
        # 生成 key=value 形式
        if isinstance(value, str):
          key_value = '''{key} = '{value}' '''.format(
                                                 key = key,
                                                 value = value
                                               )
        else:
          key_value = '''{key} = {value} '''.format(
                                              key = key,
                                              value = value
                                            )
        set_list.append(key_value)

      for key, value in where_dict.iteritems():
        if not value:
          continue
        key_value = ''
        # 生成 key=value 形式
        if isinstance(value, str):
          key_value = '''{key} = '{value}' '''.format(
                                                 key = key,
                                                 value = value
                                               )
        else:
          key_value = '''{key} = {value}'''.format(
                                              key = key,
                                              value = value
                                            )
        where_list.append(key_value)

      update_sql = '''
        UPDATE {table_name} SET {set_value} WHERE {where_value}
      '''.format(table_name  = table_name,
                 set_value   = ', '.join(set_list),
                 where_value = 'and '.join(where_list))
      return update_sql

  def create_delete_sql(self, table_name, where_dict={}):
    '''
    根据给出的列和值拼出delete语句
    '''
    if not where_dict:
      return Flase
    delete_sql = '''
      DELETE FROM {table_name}
    '''.format(table_name = table_name)
    # 将where条件的字典转化为
    where_list = []
    for item in where_dict.iteritems():
      key = item[0]
      value = item[1]
      if not value:
        continue
      key_value = ''
      # 生成 key=value 形式
      if isinstance(value, str):
        key_value = '''{key} = '{value}' '''.format(
                                               key = key,
                                               value = value
                                             )
      else:
        key_value = '''{key} = {value}'''.format(
                                            key = key,
                                            value = value
                                          )
      where_list.append(key_value)
      
    # 构造 where 后面的字符串如： and name='value'
    if where_list:
      where_str = ' and '.join(where_list)
      delete_sql += ' WHERE ' + where_str
    return delete_sql


  def execute_select_sql(self, sql_list, source_type = 'common'):
    '''
      执行select 语句 sql
    '''
    sql = ';'.join(sql_list)
    result_iter = []
    if sql_list and source_type == 'common':
      try:
        result_iter = self.cursor_select_common.execute(sql, multi=True)
      except:
        self.close_cursor('select', 'common')
        self.cursor_select_common = self.conn_common.cursor(buffered=True)
        result_iter = self.cursor_select_common.execute(sql, multi=True)
      finally:
        return result_iter
    if sql_list and source_type == 'from':
      try:
        result_iter = self.cursor_select_from.execute(sql, multi=True)
      except:
        self.close_cursor('select', 'from')
        self.cursor_select_from = self.conn_from.cursor(buffered=True)
        result_iter = self.cursor_select_from.execute(sql, multi=True)
      finally:
        return result_iter
    if sql_list and source_type == 'to':
      try:
        result_iter = self.cursor_select_to.execute(sql, multi=True)
      except:
        self.close_cursor('select', 'to')
        self.cursor_select_to = self.conn_to.cursor(buffered=True)
        result_iter = self.cursor_select_to.execute(sql, multi=True)
      finally:
        return result_iter

  def execute_dml_sql(self, sql_list=[], source_type = 'common'):
    '''
      执行dml 语句 sql
    '''
    for sql in sql_list:
      if source_type == 'common':
        self.cursor_dml_common.execute(sql)
      if source_type == 'from':
        self.cursor_dml_from.execute(sql)
      if source_type == 'to':
        self.cursor_dml_to.execute(sql)

  def get_value_iter(self, table_name, 
                     cloumn_list=[], 
                     where_dict={},
                     source_type='common'):
    '''
    获得查询的游标迭代器
    '''
    select_sql = self.create_select_sql(table_name, cloumn_list, where_dict)
    if source_type == 'common':
      return self.execute_select_sql([select_sql], 'common')
    if source_type == 'from':
      return self.execute_select_sql([select_sql], 'from')
    if source_type == 'to':
      return self.execute_select_sql([select_sql], 'to')

  def get_iter_value(self, value_iter):
    '''
    通过游标获得一条记录
    '''
    for value_i in value_iter:
      for value in value_i:
        return value

  def get_max_sharding_table_num(self):
    '''
    获得当前分表最大值
    '''
    max_sharding_table_num = None
    cloumn_list = ['system_setting_id', 'name', 'value']
    where_dict = {'name':'max_sharding_table_num'}
    system_setting_iter = self.get_value_iter('system_setting', 
                                              cloumn_list, 
                                              where_dict)
    system_setting_id, name, value = \
                       self.get_iter_value(system_setting_iter)
    return int(value)

  def get_max_sharding_db_num(self):
    '''
    获得当前分库最大值
    '''
    max_sharding_db_num = None
    cloumn_list = ['value']
    where_dict = {'name':'max_sharding_database_num'}
    system_setting_iter = self.get_value_iter('system_setting', 
                                              cloumn_list, 
                                              where_dict)
    value, = self.get_iter_value(system_setting_iter)
    return int(value)

  def get_all_sharding_tables(self):
    '''
    获得所有需要分的表
    '''
    table_names = []
    cloumn_list = ['system_setting_id', 'name', 'value']
    where_dict = {'name':'sharding_table'}
    system_setting_iter = self.get_value_iter('system_setting', 
                                              cloumn_list, 
                                              where_dict)
    for system_setting in system_setting_iter:
      for (system_setting_id, name, value) in system_setting:
        table_name = value
        table_names.append(table_name)
    return table_names

  def get_user_sharding_tables(self, user_id):
    '''
    获取用户需要迁移的表
    '''
    need_move_tables = []
    if self.is_store_owner(user_id):
      need_move_tables = self.get_system_setting_value(
                              name = 'store_owner_sharding')
    elif self.is_user_guide(user_id):
      need_move_tables = self.get_system_setting_value(
                              name = 'user_guide_sharding')
    else:
      need_move_tables = self.get_system_setting_value(
                              name = 'normal_user_sharding')
    return need_move_tables

  def is_store_owner(self, user_id):
    '''
    判断是否是店主
    '''
    cloumn_list = ['count(*)']
    where_dict = {'user_id': user_id}
    count_iter = self.get_value_iter('store', 
                                    cloumn_list, 
                                    where_dict)
    count_value, = self.get_iter_value(count_iter)
    return int(count_value)

  def is_user_guide(self, user_id):
    '''
    判断是否是导购
    '''
    cloumn_list = ['count(*)']
    where_dict = {'user_id': user_id}
    count_iter = self.get_value_iter('user_guide', 
                                    cloumn_list, 
                                    where_dict)
    count_value, = self.get_iter_value(count_iter)
    return int(count_value)

  def get_create_table_sql(self, sql, source_type='common'):
    '''
    通过 SHOW CREATA TABLE 语句获得建表语句
    '''
    if sql and source_type:
      create_table_iter = self.execute_select_sql([sql], source_type)
      table_name, create_table_sql = self.get_iter_value(create_table_iter)
      return str(create_table_sql)
      
  def create_db(self, db_config={}):
    '''
    通过指定数据库名和数据库连接描述符来创建新的数据库
    '''
    if db_config:
      self.db_config_to = db_config
      sharding.get_conn_cursor(db_config, 'to')
    db_base_name = self.get_system_setting_value('sharding_database_prefix')
    max_db_num = self.get_system_setting_value('max_sharding_database_num')
    # 通过上面获得的基本数据库名和最大数据库值来构造新创建的数据库名
    db_name = str(  db_base_name + '_' + str(int(max_db_num) + 1)  )
    create_db_sql = '''
      CREATE DATABASE {db_name}
    '''.format(db_name = db_name)
    # 执行创建数据库
    self.execute_dml_sql([create_db_sql], 'to')
    # 更新db_config_to 的配置文件
    self.db_config_to['database'] = str(db_name)
    # 从从新创建连接
    sharding.get_conn_cursor(self.db_config_to, 'to')
    # 更新系统表最大分库数max_sharding_database_num
    set_dict = {'value':int(max_db_num) + 1}
    where_dict = {'name':'max_sharding_database_num'}
    update_sql = self.create_update_sql('system_setting', 
                                        set_dict, 
                                        where_dict)
     
    self.execute_dml_sql([update_sql], 'common')
    # 向系统表插入新增的数据库名
    insert_sql = self.create_insert_sql('system_setting',
                                        ['name', 'value'],
                                        ['sharding_database', db_name])
    self.execute_dml_sql([insert_sql], 'common')
    # 向系统表中插入新建数据库连接描述符
    insert_sql = self.create_insert_sql('system_setting',
                                        ['name', 'value'],
                                        [db_name, json.dumps(self.db_config_to)])
    self.execute_dml_sql([insert_sql], 'common')
    self.commit('common')
    return db_name
   
  def create_tables(self, begin = 1, offset = 1, force=False):
    '''
    创建指定个数的表
    '''
    max_num = self.get_max_sharding_table_num()
    # 判断当前最大表个数是否大于指定的开始数和是否强制创建满足条件
    if not force and begin <= max_num:
      return
    table_names = self.get_all_sharding_tables()
    for suffix in xrange(begin, begin+offset, 1):
      table_sql_list = []
      for table_name in table_names:
        # 创建查看建表SQL语句
        show_sql = '''
          SHOW CREATE TABLE {table_name}
        '''.format(table_name  = table_name + '_' + str(suffix))
        # 获得建表语句
        create_table_sql = self.get_create_table_sql(show_sql, 
                                                 source_type = 'from')
        table_sql_list.append(create_table_sql)
      # 执行创建表的SQL
      self.execute_dml_sql(table_sql_list, source_type = 'to')

      # 生成更新当前分表最大值的SQL
      if suffix > max_num:
        update_max_sharding_num_sql = '''
          UPDATE system_setting 
          SET value={value} 
          WHERE name = 'max_sharding_table_num'
        '''.format(value=suffix)
        # 更新当前分表最大值
        self.execute_dml_sql([update_max_sharding_num_sql])
    self.commit()

  def move_data(self, username=None, db_name=None, to_table_suffix=None):
    '''
    通过用户名将其数据迁移到指定的表中
    '''
    if not username:
      return False
    if not db_name and not to_table_suffix:
      return False
    # 获得用户数据所在 数据库名 和表名
    user_id, table_flag, user_db_name = self.get_user_info(username)
    if table_flag == to_table_suffix and db_name == user_db_name:
      return False
    if not db_name:
      db_name = user_db_name
    if not to_table_suffix:
      to_table_suffix = table_flag
    # 获得用户的原数据库描述符
    db_config_from_json = self.get_system_setting_value(user_db_name)
    # 获得用户数据迁移目标数据库描述符
    db_config_to_json = self.get_system_setting_value(db_name)
    # 设置数据库链接
    self.get_conn_cursor(json.loads(db_config_from_json), 'from')
    self.get_conn_cursor(json.loads(db_config_to_json), 'to')
    # 通过user_id 获得用户需要迁移的表
    need_move_tables = self.get_user_sharding_tables(user_id)
    if not isinstance(need_move_tables, list):
      need_move_tables = [need_move_tables] 
    # 定义是每个表通过什么条件值来拆分的
    sharding_value_dict = {}
    sharding_value_dict['user_id'] = user_id
    # 获得用户导购ID
    if self.is_store_owner(user_id) or \
       self.is_user_guide(user_id):
      user_guide_id, store_id = self.get_user_guide_info(user_id, 
                                                         by='user_id')
      sharding_value_dict['user_guide_id'] = user_guide_id
      sharding_value_dict['store_id'] = store_id

    # 复制需要的表数据到指定的分表
    for table_name in need_move_tables:
      sharding_by = self.get_system_setting_value(
                         name='sharding_' + table_name + '_by')
      # 复制相关表数据到指定的分表
      self.move_table_data(
        table_from = str(table_name) + '_' + str(table_flag),
        table_to   = str(table_name) + '_' + str(to_table_suffix),
        by = sharding_by,
        value = int(sharding_value_dict[sharding_by])
      )
    # 更新表用户信息
    set_dict = {'table_flag': to_table_suffix,
                'db_name': db_name}
    where_dict = {'user_id': user_id}
    update_user_db_sql = self.create_update_sql('user', 
                                                set_dict, 
                                                where_dict)
    self.execute_dml_sql([update_user_db_sql], 'common')
    self.commit()
    # 删除原标的数据
    for table_name in need_move_tables:
      sharding_by = self.get_system_setting_value(
                         name='sharding_' + table_name + '_by')
      where_dict = {str(sharding_by):
                    int(sharding_value_dict[sharding_by])}
      self.delete_table_data(
        table_name = str(table_name) + '_' + str(table_flag),
        where_dict = where_dict
      )
    return True

  def move_table_data(self, table_from=None, table_to=None, by='id', value=None):
    '''
    实现重哪个表的数据到哪个表通过什么字段来查询
    '''
    cloumn_list = ['*']
    where_dict = {by: value}
    user_data_iter = self.get_value_iter(table_from, 
                                    cloumn_list, 
                                    where_dict,
                                    'from')
    for user_data in user_data_iter:
      # 获得表的所有字段
      columns = user_data.column_names
      # 获取所有数据并且创建新的INSERT语句执行到目标数据库
      for data in user_data:
        insert_sql = self.create_insert_sql(table_to, columns, data)
        self.execute_dml_sql([insert_sql], 'to')
      self.commit('to')

  def delete_table_data(self, table_name=None, where_dict={}):
    '''
    删除原表数据
    '''
    if not table_name or not where_dict:
      return False
    delete_data_sql = self.create_delete_sql(table_name = table_name,
                                 where_dict = where_dict)
    self.execute_dml_sql([delete_data_sql], 'from')
    self.commit()

  def get_system_setting_value(self, name=None):
    '''
    获取系统设置表的值，通过给的值名称
    '''
    cloumn_list = ['value']
    where_dict = {'name': str(name)}
    system_setting_iter = self.get_value_iter('system_setting', 
                                    cloumn_list, 
                                    where_dict)
    for system_setting in system_setting_iter:
      # 将返回的值设置为一个数组
      value_list = []
      for (value,) in system_setting:
        value_list.append(value)
      if len(value_list) == 1:
        return value_list[0]
      else:
        return value_list
    
        
  def get_user_guide_info(self, user_id=None, by=None):
    '''
    通过用户ID获得导购ID
    '''
    cloumn_list = ['user_guide_id', 'store_id']
    where_dict = {'user_id': user_id}
    user_guide_iter = self.get_value_iter('user_guide', 
                                    cloumn_list, 
                                    where_dict)
    user_guide_id, store_id = self.get_iter_value(user_guide_iter)
    return int(user_guide_id), int(store_id)
    

  def get_user_info(self, username):
    '''
    通过用户名获得用户信息
    '''
    cloumn_list = ['user_id', 'table_flag', 'db_name']
    where_dict = {'username': username}
    user_iter = self.get_value_iter('user', 
                                    cloumn_list, 
                                    where_dict)
    user_id, table_flag, db_name = \
                      self.get_iter_value(user_iter)
    return int(user_id), int(table_flag), str(db_name)
    

'''
# 分库主逻辑
if __name__=='__main__':
  # 设置默认的数据库链接参数
  db_config_common = {
    'user'    : 'root',
    'password': 'root',
    'host'    : '127.0.0.1',
    'port'    : 3306,
    'database': 'test'
  }
  # 配置用户数据所在数据库源
  db_config_from = {
    'user'    : 'root',
    'password': 'root',
    'host'    : '127.0.0.1',
    'port'    : 3306,
    'database': 'test_1'
  }
  # 配置用户数据迁移目标数据目录
  db_config_to = {
    'user'    : 'root',
    'password': 'root',
    'host'    : '127.0.0.1',
    'port'    : 3306
  }
  
  sharding = ShardingDatabase()
  # 设置公共数据库配置
  sharding.get_conn_cursor(db_config_common, 'common')
  # 设置用户原数据数据库配置
  sharding.get_conn_cursor(db_config_from, 'from')
  # 设置用户目标数据库配置
  sharding.get_conn_cursor(db_config_to, 'to')

  # 创建分库
  sharding.create_db(db_config_to)

  # 向分库中创建分表
  max_num = sharding.get_max_sharding_table_num()
  sharding.create_tables(begin = 1, offset = max_num, force=True)
'''  

# 迁移用户数据到指定的分库分表
if __name__=='__main__':
  # 设置公共库配置
  db_config_common = {
    'user'    : 'root',
    'password': 'root',
    'host'    : '127.0.0.1',
    'port'    : 3306,
    'database': 'test'
  }

  sharding = ShardingDatabase()
  # 设置公共数据库配置
  sharding.get_conn_cursor(db_config_common, 'common')
  # 指定用户数据到 哪个库 哪个表，如：用户username3数据迁移到 test_3库 10号表
  sharding.move_data('username3', 'test_3', 10)
  sharding.move_data('username7', 'test_2', 3)
  sharding.move_data('username55', 'test_2', 6)
